/*
 * Copyright (c)1998 by Angry Red Planet.
 *
 * This code is distributed under a modified form of the
 * Artistic License.  A copy of this license should have
 * been included with it; if this wasn't the case, the
 * entire package can be obtained at
 * <URL:http://www.angryredplanet.com/>.
 *
 * ----------------------------------------------------------------------
 *
 * ArpIndexedList.h
 *
 * This class represents a list whose contents can be efficiently accessed.
 * The list stores subclasses of ArpIndexedNode -- node subclasses are responsible
 * for implementing Index() and returning a value.  This list class takes the Index
 * and hashes it into a chain.  Each chain will contain up to n indexes (set by the
 * constructor).
 *
 * ----------------------------------------------------------------------
 *
 * Known Bugs
 * ~~~~~~~~~~
 *
 *	• The list does not shrink as events are removed.
 *
 * ----------------------------------------------------------------------
 *
 * History
 * ~~~~~~~
 * 11.01.98		hackborn
 * The function NodeAt(), when passed ARP_BACKWARDS, calls NodeBefore().  This
 * function seeks backwards through the list to find the first node before the time
 * supplied.  For some reason, I wrote it to respond with an error if the time
 * supplied is greater than the end of the chain.  This doesn't make sense to me
 * upon using it -- clients shouldn't need to know if a time passed in goes beyond
 * the end of the list, that situation would just be an efficient way of getting
 * the tail node.  So I've changed this to reassign an end-of-the-list time when
 * the supplied value is too large.
 *
 * 09.05.98		hackborn
 * Created this file
 */

#ifndef ARPCOLLECTIONS_ARPINDEXEDLIST_H
#define ARPCOLLECTIONS_ARPINDEXEDLIST_H


//#ifndef _LIST_H
//#include <List.h>
//#endif
#ifndef _LOCKER_H
#include <Locker.h>
#endif

#ifndef ARPCOLLECTIONS_ARPINDEXEDNODE_H
#include <ArpCollections/ArpIndexedNode.h>
#endif

//#include "Synchronization.h"


#define ARP_FORWARDS		(1)
#define ARP_BACKWARDS		(2)

// The default number of indexes that we allow in a single chain.
// See the description for the chainIndexes variable.
#define ARP_DEFAULT_CHAIN_INDEXES			(40)


class ArpIndexedList {
	friend class ArpIndexedListAccess;

	public:
		ArpIndexedList(int32 chainIndexesArg = ARP_DEFAULT_CHAIN_INDEXES);
		virtual ~ArpIndexedList();

	protected:

		// CHAIN ACCESSING
		ArpIndexedNode* HeadNode();
//		MidiListNode* HeadNode() { return(list.HeadNode()); }
		ArpIndexedNode* TailNode()				{ return tailNode; }
		virtual status_t ChainHeadFrom(int32 index, ArpIndexedNode **node);
//		MidiListNode* ChainHeadFrom(int32 index) { return(list.ChainHeadFrom(index)); }
		// Answer the index at the chain head for the supplied index.  B_ERROR is
		// returned if the index isn't in the chain.
		int32 IndexHeadFrom(int32 index);
		status_t NodeAt(int32 index, ArpIndexedNode **node,
						int32 findNode = 0, int32 direction = ARP_FORWARDS);
//		MidiListNode* FirstFrom(ulong);				// Answer the first
		status_t NodeFor(int32 index, void *rawData, ArpIndexedNode **answer);
//		MidiListNode* NodeFor(MidiEvent*);		// Answer the node wrapped around event
				// MidiListNode on or after the given ulong.
				// MidiListNode on or after the given ulong.
		int32 Count();
		int32 ChainIndexes();


// Awaiting a MIDI subclass
//		MidiEvent* Head();			// Answer the MidiEvent in my first MidiListNode
//		MidiEvent* Tail();			// Answer the MidiEvent in my last MidiListNode




		// CHAIN MANIPULATION
		status_t AddNode(ArpIndexedNode *node);
		status_t RemoveNode(ArpIndexedNode *node);
		status_t DeleteNode(ArpIndexedNode *node);
		status_t RemoveHeadNode(ArpIndexedNode **removedNode);
		status_t RemoveTailNode(ArpIndexedNode **removedNode);

		virtual void AddedNode(ArpIndexedNode *node);
		virtual void RemovingNode(ArpIndexedNode *node);
		void GrowToAtLeast(int32 countArg);		// Increase the size of chain to at least
										// the supplied value.
		void DeleteOnlyNodes();
//		void DeleteNodes();				// Delete all my midi list nodes
		// (thereby nulling out the chain), but leave the MidiEvent data
		// stored by each node untouched.		
		void DeleteNodeContents();
//		void DeleteEvents();

		// TESTING
		bool Includes(int32 index, void *rawData);
//		bool Includes(MidiEvent*);

// Awaiting a MIDI subclass
//		MidiEvent* RemoveHead();
//		virtual MidiEvent* Add(MidiEvent*);
		// AddTail() was written as a convenience for file IO, and is
		// not currently in a state that it should be used generally.
		// This won't work if any events have been added NOT via this
		// method.
//		void AddTail(MidiEvent*);
//		MidiEvent* Remove(MidiEvent*);



		// PLAYBACK -- all behaviour is undefined, that's up to subclasses.
//		virtual status_t CurrentTime(ulong*) { return(B_ERROR); };
//		virtual void SetCurrentEvent(ulong) { ; };
//		virtual MidiEvent* PlayCurrentEvent() { return(NULL); };
		// notification that playback has stopped
//		virtual void PlaybackStopped() { ; };
//		virtual void AddTo(MidiPlaybackArray*);
		
		// UTILITY
		virtual ArpIndexedList* Copy();
		virtual void CopyTo(ArpIndexedList *copy);
		virtual void ReplaceWith(ArpIndexedList *ail);
		void Print();
// Probably need these, half to wait and see.
//		void VerifyPosition(MidiEvent*);
//		void VerifyListPosition(MidiAbstractList*);

		// testing
//		virtual bool IsGeneratedList() { return(false); };

//		MidiContainer		*container;	// The container is a Version or Track.
//		MidiListNode		*currentNode;	// The event we are currently
											// on -- only used in performance.

		virtual int32 HashForIndex(int32 index);		// Answer the proper index into the
//		long IndexForPPQN(ulong);		// Answer the proper index into the
										// chain array based on the supplied time

		// FILE IO
//		void KeepChainsConsistent(MidiListNode*);

//		friend class MidiList;
		// These should be private, this is a convenience for the file IO.
		ArpIndexedNode	**chain;	// This dynamic array stores midi events
									// in spaced intervals -- the first element
									// will have 'next' elements from 00 to
									// nn, the next element will have events
									// from nn+1 to mm, etc.  When we need to
									// add higher events, we grow the chain.
		long		count;			// the size of the chain


// SHOULD BE OBSOLETE
//		virtual void SetContainer(MidiContainer*);	



	private:
//		ReadWriteLock	chainLock;
		int32			chainIndexes;	// The number of indexes we allow between chains.
										// Note the word 'index' instead of 'node.' 
										// Multiple nodes can share the same index, so the
										// actual chain size might be larger.

		ArpIndexedNode	*tmpE;		// This is just a place to hold on to a tmp
									// event without always creating a new one.
									// Originally for use for Next(), so a new
									// event for comparison isn't allocated with
									// every call to that function.  NOTE: there
									// is never any guarantee what may be in this
									// event.  Functions should ALWAYS set before
									// using.

		ArpIndexedNode	*tailNode;	// This does NOT CURRENTLY represent the tail
									// node -- I just added it as a quick hack
									// for the MidiReader.  Don't use it!

		// CHAIN ACCESSING
		status_t NodeBefore(int32 index, ArpIndexedNode **answer,
						int32 findNode);
		status_t NodeAfter(int32 index, ArpIndexedNode **answer,
						int32 findNode);

		// CHAIN MANIPULATION
		status_t pAddNode(ArpIndexedNode *node);
		status_t pRemoveNode(ArpIndexedNode *node);
		void pAddNodeAtChainHead(ArpIndexedNode *node, int32 hash);
		status_t pRemoveChainHeadAt(int32 hash, ArpIndexedNode **node);
		void pFreeChain();

		// TESTING
		bool HashExists(int32 hash);

		// UTILITY
		ArpIndexedNode* NextChainHeadFrom(int32 hash);
};

class ArpIndexedListAccess {
	public:
		enum AccessMode { AccessRead, AccessWrite };
		
		ArpIndexedListAccess(ArpIndexedList& list, AccessMode=AccessWrite);
		virtual ~ArpIndexedListAccess();

		virtual void DeleteList();

//		void SetContainer(MidiContainer *mc) { list.SetContainer(mc); }	

		// CHAIN ACCESSING
		ArpIndexedNode* HeadNode() 				{ return List()->HeadNode(); }
// Replaced by above
//		ArpIndexedNode* HeadNode() { return(list.HeadNode()); }
		ArpIndexedNode* TailNode()				{ return List()->TailNode(); }
		status_t ChainHeadFrom(int32 index, ArpIndexedNode **node)
				{ return(List()->ChainHeadFrom(index, node)); }
		// Answer the index at the chain head for the supplied index.  B_ERROR is
		// returned if the index isn't in the chain.
		int32 IndexHeadFrom(int32 index)		{ return List()->IndexHeadFrom(index); }

// Replaced by above
//		ArpIndexedNode* ChainHeadFrom(int32 index) { return(list.ChainHeadFrom(index)); }
		status_t NodeAt(int32 index, ArpIndexedNode **node,
						int32 findNode = 0, int32 direction = ARP_FORWARDS)
				{ return(List()->NodeAt(index, node, findNode, direction)); }
// Replaced by above		
//		ArpIndexedNode* FirstFrom(ulong u) { return(list.FirstFrom(u)); }
		status_t NodeFor(int32 index, void *rawData, ArpIndexedNode **answer)
				{ return(List()->NodeFor(index, rawData, answer)); }
		int32 Count() { return(List()->Count()); }

// For a MIDI subclass
		// Answer the MidiEvent in my first MidiListNode
//		MidiEvent* Head() { return(list.Head()); }
		// Answer the MidiEvent in my last MidiListNode
//		MidiEvent* Tail() { return(list.Tail()); }
//		MidiEvent* RemoveHead() { return(list.RemoveHead()); }
		// Answer the node wrapped around event
//		ArpIndexedNode* NodeFor(MidiEvent *me) { return(list.NodeFor(me)); }

		
		// CHAIN MANIPULATION
		status_t AddNode(ArpIndexedNode *node)		{ return List()->AddNode(node); }
		status_t RemoveNode(ArpIndexedNode *node)	{ return List()->RemoveNode(node); }
		status_t DeleteNode(ArpIndexedNode *node)	{ return List()->DeleteNode(node); }
		void DeleteOnlyNodes()						{ List()->DeleteOnlyNodes(); }
//		void DeleteNodes() { list.DeleteNodes(); }
		void DeleteNodeContents()					{ List()->DeleteNodeContents(); }
//		void DeleteEvents() { list.DeleteEvents(); }

		// TESTING
		bool Includes(int32 index, void *rawData)	{ return List()->Includes(index, rawData); }


// For a MIDI subclass
//		MidiEvent* Add(MidiEvent *me) { return(list.Add(me)); }
		// AddTail() was written as a convenience for file IO, and is
		// not currently in a state that it should be used generally.
		// This won't work if any events have been added NOT via this
		// method.
//		void AddTail(MidiEvent *me) { list.AddTail(me); }
//		MidiEvent* Remove(MidiEvent *me) { return(list.Remove(me)); }
//		bool Includes(MidiEvent *me) { return(list.Includes(me)); }

		// UTILITY
		ArpIndexedList* Copy()					{ return List()->Copy(); }
// Replaced by above
//		MidiList* Copy() { return(list.Copy()); }
		void ReplaceWith(ArpIndexedList *ail)	{ List()->ReplaceWith(ail); }
// Replaced by above
//		void ReplaceWith(MidiAbstractList *mal) { list.ReplaceWith(mal); }
		void Print()							{ List()->Print(); }
		virtual int32 HashForIndex(int32 index)	{ return List()->HashForIndex(index); }

		// TESTING
//		bool IsGeneratedList() { return(list.IsGeneratedList()); };


// For a MIDI subclass
		// PLAYBACK -- all behaviour is undefined, that's up to subclasses.
//		status_t CurrentTime(ulong *u) { return(list.CurrentTime(u)); }
//		void SetCurrentEvent(ulong v) { list.SetCurrentEvent(v); }
//		MidiEvent* PlayCurrentEvent() { return(list.PlayCurrentEvent()); }
		// notification that playback has stopped
//		void PlaybackStopped() { list.PlaybackStopped(); }	
		
//		void AddTo(MidiPlaybackArray *mpa) { list.AddTo(mpa); }
//		void VerifyPosition(MidiEvent *me) { list.VerifyPosition(me); }
//		void VerifyListPosition(MidiAbstractList *mal) { list.VerifyListPosition(mal); }

		
	protected:
		ArpIndexedList& list;
		AccessMode mode;

		virtual ArpIndexedList* List();
		
//		ReadLockEntry* reader;
//		WriteLockEntry* writer;
};


#endif /* ARP_INDEXED_LIST_H */
