/*
	
	ArpConfigurePanel.cpp
	
	Copyright (c)1998 by Angry Red Planet.

	This code is distributed under a modified form of the
	Artistic License.  A copy of this license should have
	been included with it; if this wasn't the case, the
	entire package can be obtained at
	<URL:http://www.angryredplanet.com/>.
*/

//#define ArpDEBUG 1
//#define USE_STREAMS 1

#ifndef _AUTOLOCK_H
#include <be/support/Autolock.h>
#endif

#ifndef ARPKERNEL_ARPCONFIGUREPANEL_H
#include <ArpKernel/ArpConfigurePanel.h>
#endif

#ifndef ARPCOLLECTIONS_ARPPTRVECTOR_H
#include <ArpCollections/ArpPtrVector.h>
#endif

#ifndef ARPCOLLECTIONS_ARPSTLVECTOR_H
#include <ArpCollections/ArpSTLVector.h>
#endif

#ifndef ARPKERNEL_ARPDEBUG_H
#include <ArpKernel/ArpDebug.h>
#endif

#ifndef _BUTTON_H
#include <be/interface/Button.h>
#endif

#ifndef _WINDOW_H
#include <be/interface/Window.h>
#endif

#include <stdlib.h>

ArpConfigurePanel::ArpConfigurePanel(BRect frame, const char* name,
								const ArpVectorI<ArpConfigurableI*>& configs,
								button_width width,
								uint32 resizingMode,
								uint32 flags)
	: BTabView(frame, name, width, resizingMode, flags),
	  mConfigs(0), mConfigViews(0), mInitSettings(0),
	  mMaxWidth(0), mMaxHeight(0), mTabWidth(0), mTabHeight(0)
{
	ArpPtrVector<BView*> getViews;
	
	const BFont* font = be_plain_font;
	
	font_height fh;
	font->GetHeight(&fh);
	const float spaceh = ceil((fh.ascent+fh.descent+fh.leading)/2);
	const float spacew = ceil(font->StringWidth("W")/2);
	
	try {
		size_t i;
		
		DB(DBALL, cdb << "Allocating vectors for " << configs.size()
					<< " configurable objects." << endl);
		
		// Initialize lists of configurable objects.
		mConfigs = new ArpPtrVector<ArpConfigurableI*>(configs.size());
		mConfigs->resize(configs.size(), 0);
		mInitSettings = new ArpSTLVector<BMessage*>;
		mInitSettings->resize(configs.size(), 0);
		for( i=0; i<configs.size(); i++ ) {
			mConfigs->at(i) = configs[i];
			mInitSettings->at(i) = new BMessage;
		}
		
		// Retrieve all configuration views.
		mConfigViews = new ArpPtrVector<BView*>;
		for( i=0; i<mConfigs->size(); i++ ) {
			mConfigs->at(i)->GetConfiguration(mInitSettings->at(i));
			getViews.resize(0);
			DB(DBALL, cdb << "Getting views for config #" << i
							<< " (" << mConfigs->at(i) << ")" << endl);
			if( mConfigs->at(i) && mConfigs->at(i)->Configure(getViews) == B_OK ) {
				DB(DBALL, cdb << "Found " << getViews.size() << endl);
				for( size_t j=0; j<getViews.size(); j++ ) {
					if( getViews[j] != 0 ) {
						mConfigViews->push_back(getViews[j]);
						getViews[j] = 0;
					}
				}
			}
		}
		
		// Set up dimensions of the tab view.
		BRect tabBound(Bounds());
		BRect tabFrame(Frame());
		mTabWidth = ceil(spacew*4 + (tabBound.Width()-tabFrame.Width()));
		mTabHeight = ceil(spaceh*4 + (tabBound.Height()-tabFrame.Height())
							+ TabHeight());
		DB(DBALL,cdb << "Tab bounds=" << tabBound
						<< ", frame=" << tabFrame << endl);
		DB(DBALL,cdb << "mTabWidth=" << mTabWidth
						<< ", mTabHeight=" << mTabHeight << endl);
						
		// Add configuration panels to the tab view.
		for( i=0; i<mConfigViews->size(); i++ ) {
			if( mConfigViews->at(i) ) {
				DB(DBALL, cdb << "Adding tab #" << i
								<< ", name="
								<< mConfigViews->at(i)->Name() << endl);
				AddTab(mConfigViews->at(i));
				mConfigViews->at(i)->MoveTo(tabBound.left+spacew*2,
											tabBound.top+spaceh*2);
				mConfigViews->at(i)->ResizeTo(tabFrame.Width()-mTabWidth,
											  tabFrame.Height()-mTabHeight);
				DB(DBALL, cdb << "This panel's frame = "
								<< mConfigViews->at(i)->Frame() << endl);
			}
		}
		
	} catch( ... ) {
		size_t i;
		for( i=0; i<getViews.size(); i++ ) delete getViews[i];
		if( mConfigViews ) {
			for( i=0; i<mConfigViews->size(); i++ ) {
				delete mConfigViews->at(i);
			}
		}
		FreeMemory();
		throw;
	}
}

ArpConfigurePanel::~ArpConfigurePanel()
{
	FreeMemory();
}

void ArpConfigurePanel::FreeMemory(void)
{
	if( mInitSettings ) {
		for( size_t i=0; i<mInitSettings->size(); i++ ) {
			delete mInitSettings->at(i);
		}
	}
	delete mInitSettings;
	mInitSettings = 0;
	delete mConfigViews;
	mConfigViews = 0;
	delete mConfigs;
	mConfigs = 0;
}

void ArpConfigurePanel::GetPreferredSize(float* width, float* height)
{
	float maxw=0, maxh=0;
	size_t i;
	
	// Get dimensions of all configuration views.
	for( i=0; i<mConfigViews->size(); i++ ) {
		if( mConfigViews->at(i) ) {
			DB(DBALL, cdb << "Processing dimens for view #" << i
							<< ", name="
							<< mConfigViews->at(i)->Name() << endl);
			float vwidth=0, vheight=0;
			mConfigViews->at(i)->GetPreferredSize(&vwidth, &vheight);
			DB(DBALL, cdb << "Preferred width=" << vwidth
							<< ", height=" << vheight << endl);
			if( vwidth > maxw ) maxw = vwidth;
			if( vheight > maxh ) maxh = vheight;
		}
	}
	
	DB(DBALL,cdb << "Final size=(" << (maxw+mTabWidth)
				<< ", " << (maxh+mTabHeight) << ")" << endl);
				
	if( width ) *width = maxw + mTabWidth + 2;
	if( height ) *height = maxh + mTabHeight + 2;
}

void ArpConfigurePanel::AttachedToWindow()
{
	DB(DBALL,cdb << "ArpConfigurePanel: Attached to window." << endl);
	inherited::AttachedToWindow();
	BView* par = Parent();
	if( par ) SetViewColor(par->ViewColor());
	else SetViewColor( ui_color(B_PANEL_BACKGROUND_COLOR) );
}

void ArpConfigurePanel::AllAttached()
{
	DB(DBALL,cdb << "ArpConfigurePanel: All attached." << endl);
	inherited::AllAttached();
	DB(DBALL,cdb << "Selecting the first tab..." << endl);
}

// ------------------------------------------------------------------------

// internal message codes
enum {
	DONE_MSG = '.don',
	REVERT_MSG = '.rev'
};

ArpConfigureView::ArpConfigureView(BRect frame, const char* name,
								const ArpVectorI<ArpConfigurableI*>& configs,
								BMessage* message,
								uint32 resizingMode,
								uint32 flags)
	: BView(frame, name, resizingMode, flags),
	  BInvoker(message, 0),
	  mPanel(0), mRevertButton(0), mDoneButton(0),
	  mPanelFrameWidth(0), mPanelFrameHeight(0),
	  mButtonsWidth(0), mButtonsHeight(0)
{
	const BFont* font = be_plain_font;
	
	font_height fh;
	font->GetHeight(&fh);
	const float spaceh = ceil((fh.ascent+fh.descent+fh.leading)/2);
	const float spacew = ceil(font->StringWidth("W")/2);
	
	mPanelFrameWidth = spacew;
	mPanelFrameHeight = spaceh;
	
	try {
		float revertWidth=0, revertHeight=0;
		mRevertButton = new BButton(frame, "revert", "Revert",
						new BMessage(REVERT_MSG), B_FOLLOW_BOTTOM|B_FOLLOW_RIGHT);
		mRevertButton->GetPreferredSize(&revertWidth,&revertHeight);
		mRevertButton->ResizeToPreferred();
		
		float doneWidth=0, doneHeight=0;
		mDoneButton = new BButton(frame, "done", "Done",
						new BMessage(DONE_MSG), B_FOLLOW_BOTTOM|B_FOLLOW_RIGHT);
		mDoneButton->GetPreferredSize(&doneWidth,&doneHeight);
		mDoneButton->ResizeToPreferred();
		
		mButtonsWidth = revertWidth+doneHeight+spacew*6;
		mButtonsHeight = revertHeight+spaceh*2;
		
		mRevertButton->MoveTo(frame.right-doneWidth-revertWidth-spacew*4,
							  frame.bottom-revertHeight-spaceh);
		mDoneButton->MoveTo(frame.right-doneWidth-spacew*2,
							frame.bottom-revertHeight-spaceh);
							
		BRect aRect(frame.left, ceil(mPanelFrameHeight/2),
					frame.right,
					frame.Height()-mButtonsHeight-mPanelFrameHeight);
		mPanel = new ArpConfigurePanel(aRect, "Panels", configs);
		
		AddChild(mPanel);
		AddChild(mRevertButton);
		AddChild(mDoneButton);
		
	} catch(...) {
		delete mRevertButton;
		delete mDoneButton;
		delete mPanel;
	}
}

ArpConfigureView::~ArpConfigureView()
{
}

void ArpConfigureView::GetPreferredSize(float* width, float* height)
{
	float pWidth=0, pHeight=0;
	mPanel->GetPreferredSize(&pWidth, &pHeight);
	//pWidth += mPanelFrameWidth;
	pHeight += mPanelFrameHeight;
	
	if( width ) *width = pWidth > mButtonsWidth ? pWidth : mButtonsWidth;
	if( height ) *height = pHeight + mButtonsHeight;
}

void ArpConfigureView::AttachedToWindow()
{
	DB(DBALL,cdb << "ArpConfigureView: Attached to window." << endl);
	inherited::AttachedToWindow();
	BView* par = Parent();
	if( par ) SetViewColor(par->ViewColor());
	else SetViewColor( ui_color(B_PANEL_BACKGROUND_COLOR) );
	mRevertButton->SetTarget(this);
	mDoneButton->SetTarget(this);
	Window()->SetDefaultButton(mDoneButton);
}

void ArpConfigureView::MessageReceived(BMessage *message)
{
	switch( message->what ) {
		case DONE_MSG: {
			Invoke();
		}
		case REVERT_MSG: {
			// not yet implemented
		}
		default: {
			inherited::MessageReceived(message);
		}
	}
}
