/*
	
	ArpLayoutable.cpp
	
	Copyright (c)1998 by Angry Red Planet.

	This code is distributed under a modified form of the
	Artistic License.  A copy of this license should have
	been included with it; if this wasn't the case, the
	entire package can be obtained at
	<URL:http://www.angryredplanet.com/>.

	The layout engine's base class, which encapsulates
	information about a rectangular UI area and how it
	can be dimensioned.
*/

//#define ArpDEBUG 1

#ifndef ARPLAYOUT_ARPLAYOUTABLE_H
#include <ArpLayout/ArpLayoutable.h>
#endif

#ifndef ARPKERNEL_ARPDEBUG_H
#include <ArpKernel/ArpDebug.h>
#endif

#ifndef _AUTOLOCK_H
#include <be/support/Autolock.h>
#endif

#include <algobase.h>
#include <float.h>
#include <string.h>

ArpMessage ArpNoParams;

static const char* NameParam = "ArpLayoutable:Name";
static const char* ChildParam = "ArpLayoutable:Child";
static const char* ConstraintsParam = "ArpLayoutable:Constraints";

const char* ArpLayoutable::BackgroundColor = "BackgroundColor";
const char* ArpLayoutable::ForegroundColor = "ForegroundColor";
const char* ArpLayoutable::BasicFont = "BasicFont";

/* ------ ArpLayoutable constructor and destructor ------
 *
 * The various ways to create and destroy ArpLayoutable
 * objects.
 */
 
ArpLayoutable::ArpLayoutable(const char* name = 0) : children(5)
{
	do_construct();
	if( name ) SetLayoutName(name);
}

ArpLayoutable::ArpLayoutable(BMessage* data) : children(5)
{
	do_construct();
	if( data ) {
		const char* newname = NULL;
		if( data->FindString(NameParam,&newname) && newname ) {
			SetLayoutName(newname);
		}
		constraints.MakeEmpty();
		data->FindMessage(ConstraintsParam,&constraints);
	}
}

void ArpLayoutable::do_construct(void)
{
	dimens_changed = true;
	inhibit_layout = true;
	focus_shown = false;
	layout_activated = false;
	parent = NULL;
	in_view = NULL;
	globals = NULL;
	PG_BackgroundColor = "StdBackColor";
	PG_ForegroundColor = "StdTextColor";
	PG_BasicFont = "BasicFont";
	PV_BackgroundColor = ui_color(B_PANEL_BACKGROUND_COLOR);
	PV_ForegroundColor.red = 0x00;
	PV_ForegroundColor.green = 0x00;
	PV_ForegroundColor.blue = 0x00;
	PV_BasicFont = *be_plain_font;
	cur_frame = cur_bounds = BRect(-99999.5,-99999.5,-99998.5,-99998.5);
	last_frame = BRect(-99997.5,-99997.5,-99996.5,-99996.5);
	error = B_NO_ERROR;
	SetLayoutName("<unnamed>");
}

ArpLayoutable::~ArpLayoutable()
{
	DB(DBALL,cdb << "Destroying ArpLayoutable: " << LayoutName() << endl);
	
	ArpLayoutable* child = 0;
	int32 index=0;
	while( (child=(ArpLayoutable*)children.ItemAt(index)) != 0 ) {
		if( !child->OwnerView() ) delete child;
		index++;
	}
}

/* ------------ ArpLayoutable archiving ------------
 *
 * Archiving and retrieving ArpLayoutable objects.
 */
ArpLayoutable* ArpLayoutable::Instantiate(BMessage* archive)
{
	if ( validate_instantiation(archive, "ArpLayoutable") ) 
		return new ArpLayoutable(archive); 
	return NULL;
}

status_t ArpLayoutable::Archive(BMessage* data, bool deep)
{
	status_t status = data->AddString("class","ArpLayoutable");
	if( status != B_NO_ERROR )
		status = data->AddMessage(ConstraintsParam,&constraints);
	if( deep ) {
		int i;
		for( i=0; i<CountLayoutChildren(); i++ ) {
			ArpLayoutable* mychild = LayoutChildAt(i);
			if( mychild ) {
				BMessage childarch;
				if( status != B_NO_ERROR )
					status = mychild->Archive(&childarch,deep);
				if( status != B_NO_ERROR )
					status = data->AddMessage(ChildParam,&childarch);
			}
		}
	}
	return status;
}

/* ------------ ArpLayoutable error control ------------
 *
 * Get and set the error state of an ArpLayoutable object.
 * These functions are safe to call on NULL objects.
 */
 
ArpLayoutable* ArpLayoutable::SetError(status_t err)
{
	if( this ) error = err;
	return this;
}

status_t ArpLayoutable::Error()
{
	if( this ) return error;
	return B_NO_MEMORY;
}

status_t ArpLayoutable::InterpretType(const ArpMessage& msg,
	const char* name, type_code type, void* data)
{
	switch(type) {
		case B_STRING_TYPE: {
			const char* str = NULL;
			status_t res = msg.FindString(name,&str);
			if( res == B_NO_ERROR ) *(ArpString*)data = str;
			return res;
		}
		case B_BOOL_TYPE:
			return msg.FindBool(name,(bool*)data);
		case B_INT8_TYPE:
			return msg.FindInt8(name,(int8*)data);
		case B_INT16_TYPE:
			return msg.FindInt16(name,(int16*)data);
		case B_INT32_TYPE:
			return msg.FindInt32(name,(int32*)data);
		case B_INT64_TYPE:
			return msg.FindInt64(name,(int64*)data);
		case B_FLOAT_TYPE:
			return msg.FindFloat(name,(float*)data);
		case B_DOUBLE_TYPE:
			return msg.FindDouble(name,(double*)data);
		case B_POINT_TYPE:
			return msg.FindPoint(name,(BPoint*)data);
		case B_RECT_TYPE:
			return msg.FindRect(name,(BRect*)data);
		case B_REF_TYPE:
			return msg.FindRef(name,(entry_ref*)data);
		case B_RGB_COLOR_TYPE:
			return msg.FindRGBColor(name,(rgb_color*)data);
		case B_PATTERN_TYPE: {
			const pattern* val;
			ssize_t size=0;
			status_t res = msg.FindData(name,type,(const void**)&val,&size);
			if( !res ) *((pattern*)data) = *val;
			return res;
		}
		case B_MESSAGE_TYPE:
			return msg.FindMessage(name,(BMessage*)data);
		case B_MESSENGER_TYPE:
			return msg.FindMessenger(name,(BMessenger*)data);
		case B_POINTER_TYPE:
			return msg.FindPointer(name,(void**)data);
		case FFont::FONT_TYPE:
			return msg.FindFont(name,(BFont*)data);
		default:
			return B_BAD_TYPE;
	}
	return B_NO_ERROR;
}

status_t ArpLayoutable::ExtractParam(
	const ArpMessage* params, const char* name, type_code type,
	const ArpMessage* globs, ArpString* gName, void* result)
{
	status_t res = B_NAME_NOT_FOUND;
	if( name && params ) {
		DB(DBALL,cdb << "Looking for param " << name << endl);
		if( params->HasData(name,type) ) {
			res = InterpretType(*params,name,type,result);
			if( gName ) *gName = (const char*)0;
			DB(DBALL,cdb << "Result = " << res << endl);
		} else if( gName ) {
			DB(DBALL,cdb << "Not found." << endl);
			const char* extname = NULL;
			extname = params->GetIndirect(name,NULL);
			if( extname ) {
				res = B_NO_ERROR;
				*gName = *extname ? extname : (const char*)0;
				DB(DBALL,cdb << "Found global reference: "
								<< gName << endl);
			}
		}
	}
	
	if( gName && !gName->IsNull() && globs && (!res || !name || !params) ) {
		status_t gres = B_NAME_NOT_FOUND;
		DB(DBALL,cdb << "Retrieving global " << *gName
					<< " for " << name << endl);
		if( globs->HasData(*gName,type) ) {
			gres = InterpretType(*globs,*gName,type,result);
			
		// If the global wasn't found, and we are being updated from the
		// complete global set, add our current value to the global params.
		} else if( globs == Globals() ) {
			DB(DBALL,cdb << "Global not found; adding." << endl);
			// This is wretched...  but fortunately, it shouldn't
			// happen very often.
			ArpMessage msg = Params(name);
			status_t res;
			type_code type;
			int32 count;
			if( (res=msg.GetInfo(name,&type,&count)) == B_OK ) {
				const void* data;
				ssize_t numBytes;
				if( (res=msg.FindData(name,type,&data,&numBytes)) == B_OK ) {
					ArpMessage glob;
					if( (res=glob.AddData(*gName,type,data,numBytes))
							== B_OK ) {
						res = AddGlobals(&glob);
					}
				}
			}
			return res;
		}
		if( !gres ) res = B_NO_ERROR;
		if( !res ) {
			DB(DBALL,cdb << "Found global value." << endl);
		}
	}
	return res;
}

/* ------------ ArpLayoutable parameters ------------
 *
 * Getting and setting object parameters.  Functions to
 * set parameters can be called on NULL objects.
 */
 
ArpLayoutable* ArpLayoutable::SetLayoutName(const char* inname)
{
	if( this ) name = inname;
	return this;
}

ArpLayoutable* ArpLayoutable::SetParams(const ArpMessage& p)
{
	if( this ) {
		DB(DBALL,cdb << "Setting parameters for "
					<< LayoutName() << ": " << p << endl);
		DoSetParams(globals,&p);
	}
	return this;
}

const ArpMessage ArpLayoutable::Params(const char* name)
{
	if( this ) {
		ArpMessage p;
		DoParams(p,name);
		DB(DBALL,cdb << "Retrieved paramters for "
					<< LayoutName() << ": " << p << endl);
		return p;
	}
	return ArpNoParams;
}

void ArpLayoutable::DoSetParams(const ArpMessage* g,
								const ArpMessage* p)
{
	DB(DBALL,cdb << "Doing set params, globals = " << g << endl);
	DB(DBALL,if( p ) cdb << "Params = " << *p << endl);
	status_t res;
	res = ExtractParam(p,BackgroundColor,B_RGB_COLOR_TYPE,
						g,&PG_BackgroundColor,&PV_BackgroundColor);
	if( !res ) {
		DB(DBALL,cdb << "Invalidating view: view color" << endl);
		if(OwnerView()) {
			OwnerView()->SetViewColor(PV_BackgroundColor);
			OwnerView()->SetLowColor(PV_BackgroundColor);
		}
		InvalidateView();
	}
	
	res = ExtractParam(p,ForegroundColor,B_RGB_COLOR_TYPE,
						g,&PG_ForegroundColor,&PV_ForegroundColor);
	if( !res ) {
		DB(DBALL,cdb << "Invalidating view: high color" << endl);
		if(OwnerView()) OwnerView()->SetHighColor(PV_ForegroundColor);
		InvalidateView();
	}
	
	res = ExtractParam(p,BasicFont,FFont::FONT_TYPE,
						g,&PG_BasicFont,&PV_BasicFont);
	if( !res ) {
		DB(DBALL,cdb << "Invalidating dimens: font" << endl);
		if(OwnerView()) OwnerView()->SetFont(&PV_BasicFont);
		InvalidateDimens();
	}
}

bool ArpLayoutable::DoParams(ArpMessage& p, const char* name)
{
	bool res = true;
	if( !name || !strcmp(name,BackgroundColor) ) {
		if( p.AddRGBColor(BackgroundColor,&PV_BackgroundColor) ) {
			res = false;
		}
		if( name ) return res;
	}
	if( !name || !strcmp(name,ForegroundColor) ) {
		if( p.AddRGBColor(ForegroundColor,&PV_ForegroundColor) ) {
			res = false;
		}
		if( name ) return res;
	}
	if( !name || !strcmp(name,BasicFont) ) {
		if( p.AddFont(BasicFont,&PV_BasicFont) ) {
			res = false;
		}
		if( name ) return res;
	}
	return res;
}

ArpLayoutable* ArpLayoutable::SetGlobals(const ArpMessage* gl)
{
	if( this ) {
		bool inhibit = inhibit_layout;
		if( !inhibit ) SetLayoutInhibit(true);
		globals = gl;
		#if ArpDEBUG
		if( gl ) {
			DB(DBALL,cdb << "Setting globals for "
						<< LayoutName() << ": " << *globals << endl);
		} else {
			DB(DBALL,cdb << "Clearing globals for "
						<< LayoutName() << "." << endl);
		}
		#endif
		DoSetParams(globals,NULL);
		int32 num = CountLayoutChildren();
		for( int32 i=0; i<num; i++ ) {
			ArpLayoutable* child = LayoutChildAt(i);
			if( child ) child->SetGlobals(globals);
		}
		if( !inhibit ) SetLayoutInhibit(false);
	}
	return this;
}

bool ArpLayoutable::RefreshGlobals(const ArpMessage* gl)
{
	if( this ) {
		if( !gl ) gl = globals;
		DB(DBALL,cdb << "Refreshing globals for "
					<< LayoutName() << " with: " << *gl << endl);
		DoSetParams(gl,NULL);
		int32 num = CountLayoutChildren();
		for( int32 i=0; i<num; i++ ) {
			ArpLayoutable* child = LayoutChildAt(i);
			if( child ) child->RefreshGlobals(gl);
		}
		return true;
	}
	return this;
}

const ArpMessage* ArpLayoutable::Globals(void)
{
	if( this ) return globals;
	return 0;
}

status_t ArpLayoutable::AddGlobals(const ArpMessage* gl)
{
	if( LayoutParent() ) return LayoutParent()->AddGlobals(gl);
	return B_NO_ERROR;
}

ArpLayoutable* ArpLayoutable::SetConstraints(const ArpMessage& c)
{
	if( this ) {
		constraints.Update(c);
		DB(DBALL,cdb << "Set constraints for "
					<< LayoutName() << ": " << constraints << endl);
		InvalidateDimens();
	}
	return this;
}

ArpMessage& ArpLayoutable::Constraints(void)
{
	if( this ) return constraints;
	return ArpNoParams;
}

void ArpLayoutable::do_inhibit_layout(bool state)
{
	inhibit_layout = state;
	int32 num = CountLayoutChildren();
	for( int32 i=0; i<num; i++ ) {
		ArpLayoutable* child = LayoutChildAt(i);
		if( child ) child->do_inhibit_layout(state);
	}
}

ArpLayoutable* ArpLayoutable::SetLayoutInhibit(bool state)
{
	if( this ) {
		DB(DBALL,cdb << "Setting layout inhibit: " << state << endl);
		inhibit_layout = state;
		ArpLayoutable* root = this;
		while( root->LayoutParent() ) {
			root = root->LayoutParent();
		}
		root->do_inhibit_layout(state);
		if( !state && dimens_changed ) {
			DB(DBALL, cdb << "dimens_changed; changing layout." << endl);
			root->LayoutChanged(false);
		}
	}
	return this;
}

void ArpLayoutable::DrawLayout(BView* inside, BRect region)
{
	int32 num = CountLayoutChildren();
	for( int32 i=0; i<num; i++ ) {
		ArpLayoutable* child = LayoutChildAt(i);
		if( child && !child->OwnerView()) child->DrawLayout(inside, region);
	}
}

/* ------------ ArpLayoutable child manipulation ------------
 *
 * Adding, examining, removing children.  This operates in
 * parallel with the BView mechanism; we need to do our own
 * implementation since we can have ArpLayoutable objects without
 * associated BView objects.
 */

BView* ArpLayoutable::FindClosestView(int32 index)
{
	ArpLayoutable* child = 0;
	
	while( (child=(ArpLayoutable*)children.ItemAt(index)) != 0 ) {
		BView* view = child->OwnerView();
		if( view ) return view;
		view = child->FindClosestView(0);
		if( view ) return view;
		index++;
	}
	
	return 0;
}

ArpLayoutable* ArpLayoutable::AddLayoutChild(ArpLayoutable* v,
											   const BMessage& c,
											   ArpLayoutable* before)
{
	// Sanity checks.
	if( !this ) return NULL;
	if( !v ) {
		SetError(B_NO_MEMORY);
		return this;
	}
	
	DB(DBALL, cdb << "Adding child: " << v->LayoutName() << endl);
	
	bool old_inhibit = LayoutInhibit();
	SetLayoutInhibit(true);
	
	// Add the given child to this ArpLayoutable.
	int bindex = -1;
	if( before ) {
		bindex = children.IndexOf(before);
		if( bindex < 0 ) {
			debugger("AddLayoutChild() failed -- the \"before\" layout and \"this\" don't match.");
			bindex = 0;
		}
		children.AddItem(v, bindex);
	} else {
		children.AddItem(v);
	}
	if( &c != &ArpNoParams ) v->constraints = c;
	v->parent = this;
	if( !(in_view=OwnerView()) && LayoutParent() ) {
		in_view = LayoutParent()->InView();
	}
	
	// Figure out the 'before' object in the BView space,
	// and add the child to our parent BView.
	BView* beforeView = 0;
	if( bindex >= 0 ) beforeView = FindClosestView(bindex+1);
	DB(DBALL, cdb << "Attaching view: parent=" << in_view
					<< ", before=" << beforeView << endl);
	v->AttachView(in_view, beforeView);
	
	// Update globals and make sure the dimensions are re-evaluated.
	if( globals && globals != v->globals ) v->SetGlobals(globals);
	DB(DBALL, cdb << "Invalidating dimens; changed="
					<< dimens_changed << endl);
	InvalidateDimens();
	
	v->SetLayoutInhibit(old_inhibit);
	
	return this;
}

bool ArpLayoutable::RemoveLayoutChild(ArpLayoutable* child)
{
	if( !child ) return false;
	
	bool old_inhibit = LayoutInhibit();
	SetLayoutInhibit(true);
	
	bool res = children.RemoveItem(child);
	DB(DBALL, cdb << "Removed child " << child->LayoutName()
					<< " from " << LayoutName() << "; result="
					<< res << endl);
	if( res ) {
		child->AttachView(0);
		child->parent = 0;
	}
	InvalidateDimens();
	
	SetLayoutInhibit(old_inhibit);
	
	return res;
}

void ArpLayoutable::AttachView(BView* par_view, BView* before)
{
	DB(DBALL, cdb << "AttachView: par_view=" << par_view
				<< ", before=" << before << ", owner=" << OwnerView()
				<< ", in=" << InView() << endl);
				
	// If this ArpLayoutable does not have an associated BView,
	// we need to run through all of its children looking for BViews
	// that instead need to be attached.
	if( !OwnerView() ) {
		in_view = par_view;
		for( int32 i=0; i<CountLayoutChildren(); i++ ) {
			ArpLayoutable* child = LayoutChildAt(i);
			if( child ) child->AttachView(par_view, before);
		}
		
	// If we are adding to a parent, do so.
	} else if( par_view ) {
		par_view->AddChild(OwnerView(), before);
	
	// If we are removing from a parent, do so.
	} else if( LayoutParent() && LayoutParent()->InView() ) {
		LayoutParent()->InView()->RemoveChild(OwnerView());
	}
}

int32 ArpLayoutable::CountLayoutChildren() const
{
	return children.CountItems();
}

ArpLayoutable* ArpLayoutable::LayoutChildAt(int32 index) const
{
	ArpLayoutable *child = (ArpLayoutable*)children.ItemAt(index);
	return child;
}

ArpLayoutable* ArpLayoutable::NextLayoutSibling()
{
	if( parent ) {
		int newindex = parent->children.IndexOf(this)+1;
		if( newindex < parent->CountLayoutChildren() ) {
			return parent->LayoutChildAt(newindex);
		}
	}
	return NULL;
}

ArpLayoutable* ArpLayoutable::PreviousLayoutSibling()
{
	if( parent ) {
		int newindex = parent->children.IndexOf(this)-1;
		if( newindex >= 0 ) {
			return parent->LayoutChildAt(newindex);
		}
	}
	return NULL;
}

bool ArpLayoutable::LayoutRemoveSelf()
{
	if( parent ) return parent->RemoveLayoutChild(this);
	return false;
}

ArpLayoutable* ArpLayoutable::FindLayoutable(const char* name)
{
	if( !name ) return 0;
	
	if( LayoutName() && strcmp(LayoutName(), name) == 0 ) return this;
	
	for( int32 i=0; i<CountLayoutChildren(); i++ ) {
		ArpLayoutable* child = LayoutChildAt(i);
		if( child ) child = child->FindLayoutable(name);
		if( child ) return child;
	}
	
	return 0;
}

void ArpLayoutable::ComputeDimens(void)
{
	DB(DBALL,cdb << "ArpLayoutable::ComputeDimens()" << endl);
	ArpLayoutable* child = (ArpLayoutable*)children.FirstItem();
	if( child ) {
		cur_dimens = child->LayoutDimens();
	} else {
		cur_dimens.min_width = cur_dimens.min_height
			= cur_dimens.pref_width = cur_dimens.pref_height = 0;
		cur_dimens.max_width = cur_dimens.max_height = FLT_MAX/2;
	}
}

const ArpLayoutDimens& ArpLayoutable::LayoutDimens(void)
{
	if( dimens_changed ) {
		ComputeDimens();
		DB(DBALL,cdb << "Uncorrected dimensions for "
					<< LayoutName() << endl);
		DB(DBALL,cdb << "  min  width=" << cur_dimens.min_width
					<< " height=" << cur_dimens.min_height << endl);
		DB(DBALL,cdb << "  pref width=" << cur_dimens.pref_width
					<< " height=" << cur_dimens.pref_height << endl);
		DB(DBALL,cdb << "  max  width=" << cur_dimens.max_width
					<< " height=" << cur_dimens.max_height << endl);
		cur_dimens.min_width = ceil(max(cur_dimens.min_width,(float)0));
		cur_dimens.min_height = ceil(max(cur_dimens.min_height,(float)0));
		cur_dimens.pref_width = ceil(max(cur_dimens.min_width,
										cur_dimens.pref_width));
		cur_dimens.pref_height = ceil(max(cur_dimens.min_height,
										cur_dimens.pref_height));
		cur_dimens.max_width = floor(max(cur_dimens.max_width,(float)0));
		cur_dimens.max_height = floor(max(cur_dimens.max_height,(float)0));
		dimens_changed = false;
	}
	return cur_dimens;
}

void ArpLayoutable::ResizeLayout(float width, float height,
								  bool force)
{
	DB(DBALL,cdb << "ArpLayoutable::LayoutResize(" << width
				 << "," << height << "): " << LayoutName() << endl);
	cur_frame.right = cur_frame.left + width;
	cur_frame.bottom = cur_frame.top + height;
	LayoutChanged(force);
}

void ArpLayoutable::MoveLayout(float left, float top, bool force)
{
	DB(DBALL,cdb << "ArpLayoutable::LayoutMove(" << left
				 << "," << top << "): " << LayoutName() << endl);
	cur_frame.OffsetTo(left,top);
	LayoutChanged(force);
}

void ArpLayoutable::SetLayout(const BRect& frame, bool force)
{
	DB(DBALL,cdb << "ArpLayoutable::SetLayout(" << frame << "): "
				  << LayoutName() << endl);
	cur_frame = frame;
	LayoutChanged(force);
}

void ArpLayoutable::RequestLayout(bool force)
{
	DB(DBALL,cdb << "ArpLayoutable::RequestLayout()" << endl);
	LayoutChanged(force);
}

void ArpLayoutable::InvalidateDimens(void)
{
	if( dimens_changed ) return;
	ArpLayoutable* root = this;
	ArpLayoutable* last = this;
	while( root ) {
		DB(DBALL,cdb << "Invalidating dimensions for: "
					<< root->LayoutName() << endl);
		root->dimens_changed = true;
		root->last_frame.left = root->last_frame.top = -1000;
		root->last_frame.right = root->last_frame.bottom = -2000;
		root->InvalidateView();
		last = root;
		root = root->LayoutParent();
	}
	if( !inhibit_layout && last ) last->LayoutChanged(false);
}

void ArpLayoutable::InvalidateView(void)
{
	//printf("Invalidate view %s: in=%x, owner=%x\n",
	//		LayoutName(), InView(), OwnerView());
	if( InView() ) InView()->Invalidate();
}

void ArpLayoutable::SetFocusShown(bool state, bool andParent)
{
	focus_shown = state;
	if( andParent && LayoutParent() ) LayoutParent()->SetFocusShown(state);
}

bool ArpLayoutable::FocusShown(void) const
{
	return focus_shown;
}

void ArpLayoutable::SetLayoutActivated(bool state)
{
	layout_activated = state;
	int32 num = CountLayoutChildren();
	for( int32 i=0; i<num; i++ ) {
		ArpLayoutable* child = LayoutChildAt(i);
		if( child ) child->SetLayoutActivated(state);
	}
}

bool ArpLayoutable::LayoutActivated(void) const
{
	return layout_activated;
}

void ArpLayoutable::LayoutChanged(bool force)
{
	DB(DBALL,cdb << "ArpLayoutable::LayoutChanged(" << force
				<< ")" << ": " << LayoutName() << endl);
	//ArpLayoutable* parent = LayoutParent();
	BView* view = OwnerView();
	bool changed = false;
	
	// Make sure this object's dimension information has been
	// computed by requesting it.
	LayoutDimens();
	
	// Special case for the root of a layout hierarchy -- we don't
	// tell it where to go; it tells us.
	if( IsLayoutRoot() ) {
		if( force || last_frame != cur_frame ) {
			last_frame = cur_frame;
			cur_bounds = cur_frame;
			if( view ) cur_bounds.OffsetTo(0,0);
			Layout();
		}
		return;
	}
	
	// Compute the interior drawing frame of this object.
	// If there is no BView associated with it, this is simply
	// the real offset we were given -- we use the same
	// coordinate system as our parent.
	// Otherwise, we have our own coordinate system and the
	// inner frame is simply placed at (0,0).
	cur_bounds = cur_frame;
	if( view ) cur_bounds.OffsetTo(0,0);
	
	//printf("ArpLayoutable::LayoutChanged() of %s\n",LayoutName());
	if( force || cur_frame.left != last_frame.left
			  || cur_frame.top != last_frame.top ) {
		DB(DBALL,cdb << "The frame position has changed..." << endl);
		if( view ) {
			DB(DBALL,cdb << "ArpLayoutable::SetFrame "
						<< LayoutName() << " -- moving." << endl);
			//printf("Move %s to (%f,%f)\n",LayoutName(),
			//	cur_frame.left,cur_frame.top);
			view->MoveTo(cur_frame.left, cur_frame.top);
		} else {
			// Only need to re-layout children if we have no
			// BView, so their own offsets will change.
			changed = true;
		}
	}
	if( force || cur_frame.Width() != last_frame.Width()
			  || cur_frame.Height() != last_frame.Height() ) {
		DB(DBALL,cdb << "The frame size has changed..." << endl);
		changed = true;
		if( view ) {
			DB(DBALL,cdb << "ArpLayoutable::SetFrame "
						<< LayoutName() << " -- resizing." << endl);
			//printf("Resize %s to (%f-%f)\n",LayoutName(),
			//	cur_frame.Width(),cur_frame.Height());
			view->ResizeTo(cur_frame.Width(), cur_frame.Height());
		}
	}
	last_frame = cur_frame;
	if( changed ) {
		Layout();
	}
}

void ArpLayoutable::Layout()
{
	BRect frm = LayoutBounds();
	DB(DBALL,cdb << "ArpLayoutable::Layout() -- " << LayoutName()
			<< " " << frm << endl);
	ArpLayoutable* child = (ArpLayoutable*)children.FirstItem();
	if( child ) {
		child->SetLayout(frm);
	}
}
